import tidypolars as tp
from tidypolars import col
import polars as pl

def _repeat(x, times):
    if not isinstance(x, list):
        x = [x]
    return x * times

def test_arrange1():
    """Can arrange ascending"""
    df = tp.Tibble(x = ['a', 'a', 'b'], y = [2, 1, 3])
    actual = df.arrange('y')
    expected = tp.Tibble(x = ['a', 'a', 'b'], y = [1, 2, 3])
    assert actual.frame_equal(expected), "arrange ascending failed"

def test_arrange2():
    """Can arrange descending"""
    df = tp.Tibble({'x': ['a', 'a', 'b'], 'y': [2, 1, 3]})
    actual = df.arrange('x', 'y', desc = [True, False])
    expected = tp.Tibble({'x': ['b', 'a', 'a'], 'y': [3, 1, 2]})
    assert actual.frame_equal(expected), "arrange descending failed"

def test_bind_cols():
    """Can bind_cols"""
    df1 = tp.Tibble({'x': ['a', 'a', 'b'], 'y': [1, 2, 3]})
    df2 = tp.Tibble({'z': [4, 4, 4]})
    actual = df1.bind_cols(df2)
    expected = tp.Tibble({'x': ['a', 'a', 'b'], 'y': [1, 2, 3], 'z':[4, 4, 4]})
    assert actual.frame_equal(expected), "bind cols failed"

def test_bind_rows():
    """Can bind rows"""
    df1 = tp.Tibble({'x': ['a', 'a'], 'y': [2, 1]})
    df2 = tp.Tibble({'x': ['b'], 'y': [3]})
    actual = df1.bind_rows(df2)
    expected = tp.Tibble({'x': ['a', 'a', 'b'], 'y': [2, 1, 3]})
    assert actual.frame_equal(expected), "bind rows failed"

def test_distinct_empty():
    """Can distinct columns"""
    df = tp.Tibble({'x': ['a', 'a', 'b'], 'y': ['a', 'a', 'b']})
    actual = df.distinct()
    expected = tp.Tibble({'x': ['a', 'b'], 'y': ['a', 'b']})
    assert actual.frame_equal(expected), "empty distinct failed"

def test_distinct_select():
    """Can distinct columns"""
    df = tp.Tibble({'x': ['a', 'a', 'b'], 'y': [2, 1, 3]})
    actual = df.distinct('x')
    expected = tp.Tibble({'x': ['a', 'b']})
    assert actual.frame_equal(expected), "distinct with select failed"

def test_fill():
    """Can fill"""
    df = tp.Tibble({'chr': ['a', None], 'int': [1, None]})
    actual = df.fill('chr', 'int')
    expected = tp.Tibble({'chr': ['a', 'a'], 'int': [1, 1]})
    assert actual.frame_equal(expected), "fill failed"

def test_filter():
    """Can filter multiple conditions"""
    df = tp.Tibble({'x': range(10), 'y': range(10)})
    actual = df.filter(col('x') <= 3, col('y') < 2)
    expected = tp.Tibble({'x': range(2), 'y': range(2)})
    assert actual.frame_equal(expected), "filter failed"

def test_mutate():
    """Can edit existing columns and can add columns"""
    df = tp.Tibble({'x': _repeat(1, 3), 'y': _repeat(2, 3)})
    actual = df.mutate(double_x = col('x') * 2,
                       y = col('y') + 10)
    expected = tp.Tibble(
        {'x': _repeat(1, 3),
         'y': _repeat(12, 3),
          'double_x': _repeat(2, 3)}
    )
    assert actual.frame_equal(expected), "mutate failed"

def test_mutate_across():
    """Can mutate multiple columns simultaneously"""
    df = tp.Tibble({'x': _repeat(1, 3), 'y': _repeat(2, 3)})
    actual = df.mutate(col(['x', 'y']) * 2,
                       x_plus_y = col('x') + col('y'))
    expected = tp.Tibble(
        {'x': _repeat(2, 3),
         'y': _repeat(4, 3),
         'x_plus_y': _repeat(3, 3)}
    )
    assert actual.frame_equal(expected), "mutate across failed"

def test_names():
    """Can get column names"""
    df = tp.Tibble({'x': _repeat(1, 3), 'y': _repeat(2, 3)})
    assert df.names == ['x', 'y'], "names failed"

def test_ncol():
    """Can number of columns"""
    df = tp.Tibble({'x': _repeat(1, 3), 'y': _repeat(2, 3)})
    assert df.ncol == 2, "ncol failed"

def test_nrow():
    """Can number of rows"""
    df = tp.Tibble({'x': _repeat(1, 3), 'y': _repeat(2, 3)})
    assert df.nrow == 3, "nrow failed"

def test_pivot_longer1():
    "Can pivot all (unspecified) cols to long"
    df = tp.Tibble({'x': [1, 2], 'y': [3, 4]})
    actual = df.pivot_longer()
    expected = tp.Tibble({'name': ['x', 'x', 'y', 'y'], 'value': range(1, 5)})
    assert actual.frame_equal(expected), "unspecified pivot_longer failed"

def test_pivot_longer2():
    """Can pivot all (specified) cols to long"""
    df = tp.Tibble({'x': [1, 2], 'y': [3, 4]})
    actual = df.pivot_longer(['x', 'y'])
    expected = tp.Tibble({'name': ['x', 'x', 'y', 'y'], 'value': range(1, 5)})
    assert actual.frame_equal(expected), "specified pivot_longer failed"

def test_pivot_wider1():
    """Can pivot all cols to wide"""
    df = tp.Tibble({'label': ['x', 'y', 'z'], 'val': range(1, 4)})
    actual = df.pivot_wider(names_from = 'label', values_from = 'val').select('x', 'y', 'z')
    expected = tp.Tibble({'x': [1], 'y': [2], 'z': [3]})
    assert actual.frame_equal(expected), "pivot_wider all cols failed"

def test_pivot_wider2():
    """Can pivot cols to wide with id col"""
    df = tp.Tibble({'id': _repeat(1, 3), 'label': ['x', 'y', 'z'], 'val': range(1, 4)})
    actual = df.pivot_wider(names_from = 'label', values_from = 'val').select('id', 'x', 'y', 'z')
    expected = tp.Tibble({'id': [1], 'x': [1], 'y': [2], 'z': [3]})
    assert actual.frame_equal(expected), "pivot_wider with id failed"

def test_pivot_wider3():
    """Can pivot cols to wide with values filled"""
    df = tp.Tibble({'id': _repeat(1, 3), 'label': ['x', 'y', 'z'], 'val': range(1, 4)})
    actual = (
        df.pivot_wider(names_from = 'label', values_from = 'id', values_fill = 0)
        .select('val', 'x', 'y', 'z').arrange('val')
    )
    expected = tp.Tibble({'val': [1, 2, 3], 'x': [1, 0, 0], 'y': [0, 1, 0], 'z': [0, 0, 1]})
    assert actual.frame_equal(expected), "pivot_wider with values filled failed"

def test_pull():
    """Can use pull"""
    df = tp.Tibble({'x': _repeat(1, 3), 'y': _repeat(2, 3)})
    actual = df.pull('x')
    expected = df.to_polars().get_column('x')
    assert actual == expected, "pull failed"

def test_relocate1():
    """Can relocate before columns"""
    df = tp.Tibble({'x': range(3), 'y': range(3), 'z': range(3)})
    actual = df.relocate('y', 'z', before = 'x')
    expected = df.select('y', 'z', 'x')
    assert actual.frame_equal(expected), "relocate before failed"

def test_relocate2():
    """Can relocate after columns"""
    df = tp.Tibble({'x': range(3), 'y': range(3), 'z': range(3)})
    actual = df.relocate('z', 'y', after = 'x')
    expected = df.select('x', 'z', 'y')
    assert actual.frame_equal(expected), "relocate after failed"

def test_rename_dplyr_kwargs():
    """Can rename - dplyr interface (kwargs)"""
    df = tp.Tibble({'x': range(3), 'y': range(3), 'z': range(3)})
    actual = df.rename(new_x = 'x', new_y = 'y')
    expected = tp.Tibble({'new_x': range(3), 'new_y': range(3), 'z': range(3)})
    assert actual.frame_equal(expected), "dplyr rename failed"

def test_rename_dplyr_strings():
    """Can rename - dplyr interface (strings)"""
    df = tp.Tibble({'x': range(3), 'y': range(3), 'z': range(3)})
    actual = df.rename('new_x', 'x', 'new_y', 'y')
    expected = tp.Tibble({'new_x': range(3), 'new_y': range(3), 'z': range(3)})
    assert actual.frame_equal(expected), "dplyr rename failed"

def test_rename_pandas():
    """Can rename - pandas interface"""
    df = tp.Tibble({'x': range(3), 'y': range(3), 'z': range(3)})
    actual = df.rename({'x': 'new_x', 'y': 'new_y'})
    expected = tp.Tibble({'new_x': range(3), 'y': range(3), 'z': range(3)})
    assert actual.frame_equal(expected), "pandas rename failed"

def test_select():
    """Can select columns"""
    df = tp.Tibble({'x': range(3), 'y': range(3), 'z': range(3)})
    actual = df.select('x', 'z')
    expected = df[['x', 'z']]
    assert actual.frame_equal(expected), "select failed"

def test_slice():
    """Can slice"""
    df = tp.Tibble({'x': range(3), 'y': ['a', 'a', 'b']})
    actual = df.slice(0, 2)
    expected = tp.Tibble({'x': [0, 2], 'y': ['a', 'b']})
    assert actual.frame_equal(expected), "slice failed"

def test_slice_head():
    """Can slice_head"""
    df = tp.Tibble({'x': range(3), 'y': ['a', 'a', 'b']})
    actual = df.slice_head(2)
    expected = tp.Tibble({'x': [0, 1], 'y': ['a', 'a']})
    assert actual.frame_equal(expected), "slice_head failed"

def test_slice_tail():
    """Can slice_tail by group"""
    df = tp.Tibble({'x': range(3), 'y': ['a', 'a', 'b']})
    actual = df.slice_tail(2)
    expected = tp.Tibble({'x': [1, 2], 'y': ['a', 'b']})
    assert actual.frame_equal(expected), "slice_tail failed"

def test_summarize():
    """Can use summarize"""
    df = tp.Tibble({'x': range(3), 'y': range(3), 'z': range(3)})
    actual = df.summarize(avg_x = col('x').mean())
    expected = tp.Tibble({'avg_x': [1]})
    assert actual.frame_equal(expected), "ungrouped summarize failed"

def test_summarize_across():
    """Can use summarize_across"""
    df = tp.Tibble(x = range(3), y = range(3), z = range(3))
    actual = df.summarize(col(['x', 'y']).max().prefix('max_'),
                          avg_x = col('x').mean())
    expected = tp.Tibble({'max_x': [2], 'max_y': [2], 'avg_x': [1]})
    assert actual.frame_equal(expected), "ungrouped summarize across failed"

def test_to_polars():
    """Can convert to a polars DataFrame"""
    df = tp.Tibble({'x': range(3), 'y': range(3), 'z': range(3)})
    assert isinstance(df.to_polars(), pl.DataFrame), "to_polars failed"

def test_funs_in_a_row():
    df = tp.Tibble(x = range(3), y = range(3), z = range(3))
    df.distinct()
    df.drop('x')
    df.filter(col('x') < 7)
    df.head()
    df.mutate(col('x') * 2)
    df.relocate('y', before = 'x')
    df.rename({'x': 'new_x'})
    df.select('x', 'y')
    df.slice(1)
    df.slice_head()
    df.slice_tail()
    df.tail()
    df.arrange('x', 'y')
    assert True, "Functions in a row failed"
